import os
import sys
import boto3
import logging
import argparse
import sagemaker
from sagemaker import Model
from sagemaker.utils import name_from_base
from pathlib import Path

root = logging.getLogger()
if root.handlers:
    for handler in root.handlers:
        root.removeHandler(handler)
logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

if __name__ == "__main__":
    
    # Define and parse command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--device",
        type=str,
        help="Device name, allowed values \"inf2\" or \"gpu\""
    )
    parser.add_argument(
        "--aws-region",
        type=str,
        default="us-east-1",
        help="AWS region, default=\"us-east-1\"",
    )

    parser.add_argument(
        "--role-arn",
        type=str,
        help="ARN for the role to be used to deploy the model",
    )

    parser.add_argument(
        "--bucket",
        type=str,
        help="S3 bucket name (without s3://) that contains the model binaries",
    )

    parser.add_argument(
        "--model-id",
        type=str,
        help="Model id",
    )

    parser.add_argument(
        "--prefix",
        type=str,
        default="lmi",
        help="S3 bucket prefix in which model binaries are kept",
    )

    parser.add_argument(
        "--inf2-instance-type",
        type=str,
        default="ml.inf2.24xlarge",
        help="Inf2 instance type for deploying the model",
    )

    parser.add_argument(
        "--inf2-image-uri",
        type=str,
        default="763104351884.dkr.ecr.us-east-1.amazonaws.com/djl-inference:0.27.0-neuronx-sdk2.18.1",
        help="Image URI for the Inf2 inference container",
    )

    parser.add_argument(
        "--model-s3-uri",
        type=str,
        help="S3 URI for model.tar.gzr",
    )

    parser.add_argument(
        "--neuronx-artifacts-s3-uri",
        type=str,
        help="S3 URI for Neuronx artifacts",
    )
    parser.add_argument(
        "--script-path",
        type=str,
        help="path to this script directory",
    )
    parser.add_argument(
        "--initial-instance-count",
        type=int,
        help="number of instances to start with",
    )
    parser.add_argument(
        "--model-data-timeout",
        type=int,
        help="increase the timeout to download large model",
    )

    args = parser.parse_args()
    logger.info(f"args={args}")


    dev = args.device
    aws_region = args.aws_region
    os.environ['AWS_DEFAULT_REGION'] = aws_region
    role = args.role_arn
    bucket_name = args.bucket
    model_id = args.model_id
    prefix = args.prefix
    s3_uri = args.model_s3_uri
    neuronx_artifacts = args.neuronx_artifacts_s3_uri
    script_path = args.script_path
    if dev == 'gpu':        
        instance_type = args.gpu_instance_type
        image_uri = args.gpu_image_uri
    elif dev == 'inf2':
        instance_type = args.inf2_instance_type
        image_uri = args.inf2_image_uri
    else:
        logger.error('Invalid device type')
        sys.exit(-1)

    model_name = f"{model_id}-{dev}".replace("/", "-")
    endpoint_name = sagemaker.utils.name_from_base(model_name).replace(".", "-")
    logger.info(f"going to deploy model_id={model_id}, endpoint={endpoint_name},\
                  s3_uri={s3_uri}\
                  aws_region={aws_region}\
                  role={role},\
                  bucket_name={bucket_name},\
                  prefix={prefix},\
                  instance_type={instance_type},\
                  image_uri={image_uri}")

    boto3_session=boto3.session.Session(region_name=aws_region)
    smr = boto3.client('sagemaker-runtime')
    sm = boto3.client('sagemaker')
    # sagemaker session for interacting with different AWS APIs
    sess = sagemaker.session.Session(boto3_session, 
                                    sagemaker_client=sm, 
                                    sagemaker_runtime_client=smr)
        
    logger.info(f'Deploying on {dev}')
    logger.info("======================================")
    logger.info(f'Will load artifacts from {s3_uri}')
    logger.info("======================================")

    logger.info("======================================")
    logger.info(f'Using Container image {image_uri}')
    logger.info("======================================")
    model = Model(
        name=endpoint_name,
        # Enable SageMaker uncompressed model artifacts
        model_data={
            "S3DataSource": {
                    "S3Uri": s3_uri,
                    "S3DataType": "S3Prefix",
                    "CompressionType": "Gzip",
            }
        },
        image_uri=image_uri,
        role=role,
        env = {
            "NEURON_COMPILE_CACHE_URL": neuronx_artifacts
        },
        sagemaker_session=sess
        #env - set TS_INSTALL_PY_DEP_PER_MODEL to true, if you are using Pytorch serving
        #this will tell server to run requirements.txt to deploy any additional packages
    )
    logger.info(model)

    # Path to the endpoint.txt file in the higher directory
    endpoint_file_path = os.path.join(script_path, "endpoint.txt")
    logger.info(f'endpoint_file_path={endpoint_file_path}')

    logger.info(f'\nModel deployment initiated on {dev}\nEndpoint Name: {endpoint_name}\n')
    if "trn1" in instance_type:
        model.deploy(
            initial_instance_count=1,
            instance_type=instance_type,
            endpoint_name=endpoint_name,
            model_data_download_timeout=2400,  # increase the timeout to download large model
            container_startup_health_check_timeout=2400,  # increase the timeout to load large model,
            wait=True,
        )
    else:
        model.deploy(
            initial_instance_count=1,
            instance_type=instance_type,
            endpoint_name=endpoint_name,
            volume_size=512, # not allowed for the selected Instance type ml.g5.12xlarge and ml.trn1.32xlarge
            model_data_download_timeout=2400,  # increase the timeout to download large model
            container_startup_health_check_timeout=2400,  # increase the timeout to load large model,
            wait=True,
        )   
    logger.info(f'Model deployment on {dev}\nEndpoint Name: {endpoint_name} finished\n')
    logger.info(f'Now writing the endpoint and end of DEPLOY')
    # Write the endpoint name to the endpoint.txt file in the higher directory
    Path(endpoint_file_path).write_text(endpoint_name)
    
 


"""
Collects and logs GPU and CPU metrics to an EC2 metrics CSV file. This file is populated during the
duration of the inferences against the model deployed on the EC2 instance.
"""
import csv
import time
import psutil
import logging
from fmbench.scripts import constants
from nvitop import Device, ResourceMetricCollector

# Setup logging
logging.basicConfig(
    format="[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)

# Global flag to control data collection for now, change in future.
collecting = True


def stop_collect(collector=None):
    """
    Stops the data collection process by setting the global flag 'collecting' to False.
    """
    global collecting
    collecting = False
    logger.info("Stopped collection")


def _collect_ec2_utilization_metrics():
    """
    Starts the data collection process by initializing the ResourceMetricCollector and collecting metrics at regular intervals.
    """
    global collecting
    logger.info("Starting collection")

    def on_collect(metrics):
        """
        Collects GPU and CPU metrics, then appends them to the CSV file.

        Parameters:
        - metrics: The collected metrics from the ResourceMetricCollector.

        Returns:
        - bool: Returns False if the collection is stopped, otherwise True.
        """
        if not collecting:
            return False

        try:
            # Open the CSV file in append mode and write the collected metrics
            with open(constants.EC2_SYSTEM_METRICS_FNAME, mode="a", newline="") as csv_file:
                csv_writer = csv.writer(csv_file)

                # Collect CPU mean utilization
                cpu_percent_mean = metrics.get(
                    "metrics-daemon/host/cpu_percent (%)/mean", psutil.cpu_percent()
                )
                memory_percent_mean = metrics.get(
                    "metrics-daemon/host/memory_percent (%)/mean",
                    psutil.virtual_memory().percent,
                )
                memory_used_mean = metrics.get(
                    "metrics-daemon/host/memory_used (GiB)/mean",
                    psutil.virtual_memory().available,
                )

                # Extract the current timestamp
                timestamp = time.strftime("%Y-%m-%d %H:%M:%S")

                # Initialize variables to sum GPU metrics
                total_gpu_utilization = 0
                total_gpu_memory_used = 0
                total_gpu_memory_free = 0
                total_gpu_memory_total = 0

                num_gpus = len(Device.cuda.all())
                # logger.info(f"Number of GPUs detected: {num_gpus}")

                # Iterate over all detected GPUs
                for gpu_id in range(num_gpus):
                    gpu_utilization_mean = metrics.get(
                        f"metrics-daemon/cuda:{gpu_id} (gpu:{gpu_id})/gpu_utilization (%)/mean",
                        None,
                    )
                    gpu_memory_used_mean = metrics.get(
                        f"metrics-daemon/cuda:{gpu_id} (gpu:{gpu_id})/memory_used (MiB)/mean",
                        None,
                    )
                    gpu_memory_free_mean = metrics.get(
                        f"metrics-daemon/cuda:{gpu_id} (gpu:{gpu_id})/memory_free (MiB)/mean",
                        None,
                    )
                    gpu_memory_total_mean = metrics.get(
                        f"metrics-daemon/cuda:{gpu_id} (gpu:{gpu_id})/memory_total (MiB)/mean",
                        None,
                    )

                    if gpu_utilization_mean is not None:
                        total_gpu_utilization += gpu_utilization_mean
                    if gpu_memory_used_mean is not None:
                        total_gpu_memory_used += gpu_memory_used_mean
                    if gpu_memory_free_mean is not None:
                        total_gpu_memory_free += gpu_memory_free_mean
                    if gpu_memory_total_mean is not None:
                        total_gpu_memory_total += gpu_memory_total_mean

                # Calculate the mean values across all GPUs
                gpu_utilization_mean_total = (
                    total_gpu_utilization / num_gpus if num_gpus > 0 else None
                )
                gpu_memory_used_mean_total = (
                    total_gpu_memory_used / num_gpus if num_gpus > 0 else None
                )
                gpu_memory_free_mean_total = (
                    total_gpu_memory_free / num_gpus if num_gpus > 0 else None
                )
                gpu_memory_total_mean_total = (
                    total_gpu_memory_total / num_gpus if num_gpus > 0 else None
                )

                # Write the row to the CSV file
                row = [
                    timestamp,
                    cpu_percent_mean,
                    memory_percent_mean,
                    memory_used_mean,
                    gpu_utilization_mean_total,
                    gpu_memory_used_mean_total,
                    gpu_memory_free_mean_total,
                    gpu_memory_total_mean_total,
                ]
                # logger.info(f"Writing row: {row}")
                csv_writer.writerow(row)

        except ValueError as e:
            return False

        return True

    # Start the collector and run in the background
    collector = ResourceMetricCollector(Device.cuda.all())
    logger.info("Starting Daemon collector to run in background")
    collector.daemonize(
        on_collect,
        interval=constants.EC2_UTILIZATION_METRICS_INTERVAL,
        on_stop=stop_collect,
    )


def collect_ec2_metrics():
    """
    Initializes the CSV file with headers and starts the metrics collection process.
    """
    global collecting
    collecting = True
    # Initialize the CSV file and write the header once
    with open(constants.EC2_SYSTEM_METRICS_FNAME, mode="w", newline="") as csv_file:
        csv_writer = csv.writer(csv_file)
        header = [
            "timestamp",
            "cpu_percent_mean",
            "memory_percent_mean",
            "memory_used_mean",
            "gpu_utilization_mean",
            "gpu_memory_used_mean",
            "gpu_memory_free_mean",
            "gpu_memory_total_mean",
        ]
        logger.info(f"Writing header: {header}")
        csv_writer.writerow(header)

    # Call the function to start collecting metrics
    _collect_ec2_utilization_metrics()



import os
import json
import copy
import time
import stat
import logging
import requests
import tempfile
import subprocess
import pandas as pd
from pathlib import Path
from datetime import datetime
from typing import Dict, Optional
from fmbench.scripts import constants
from fmbench.utils import count_tokens
from fmbench.scripts.inference_containers.utils import get_accelerator_type
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)
from fmbench.scripts.inference_containers.utils import (STOP_AND_RM_CONTAINER,
                                                        FMBENCH_MODEL_CONTAINER_NAME)
                                            
# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ollamaPredictor(FMBenchPredictor):
    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict], 
                 metadata: Optional[Dict]):
        try:
            self._endpoint_name: str = endpoint_name
            self._inference_spec: Dict = inference_spec
            self._accelerator = get_accelerator_type()  # Not needed for Ollama, but keeping in case it's useful
        except Exception as e:
            logger.error(f"create_predictor, exception occurred while creating predictor "
                         f"for endpoint_name={self._endpoint_name}, exception={e}")
        logger.info(f"_endpoint_name={self._endpoint_name}, _inference_spec={self._inference_spec}")

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        response_json: Optional[Dict] = None
        response: Optional[str] = None
        latency: Optional[float] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None
        prompt_tokens: Optional[int] = None
        completion_tokens: Optional[int] = None

        # Get the prompt for the Ollama endpoint
        prompt: str = payload['inputs']
        # Calculate the number of tokens in the prompt
        prompt_tokens = count_tokens(payload['inputs'])

        try:
            # Adjust the payload for Ollama
            payload2 = copy.deepcopy(payload)
            payload2['prompt'] = payload2.pop('inputs')  # Replace 'inputs' with 'prompt'
            payload2 = payload2 | self._inference_spec.get("parameters", {})  # Add inference parameters if present

            # Send POST request to the Ollama endpoint
            st = time.perf_counter()  # Start latency timer
            response = requests.post(self._endpoint_name, json=payload2)
            latency = time.perf_counter() - st  # Record the latency

            # Raise exception if the response failed
            response.raise_for_status()
            full_output = response.text

            # Extract the generated response (removing the original prompt if necessary)
            answer_only = full_output.replace(prompt, "", 1).strip('["]?\n')
            response_json = dict(generated_text=answer_only)

            # Count the completion tokens for the generated text
            completion_tokens = count_tokens(response_json.get("generated_text"))

        except Exception as e:
            logger.error(f"get_prediction, exception occurred while getting prediction for payload={payload} "
                         f"from predictor={self._endpoint_name}, response={response}, exception={e}")

        return FMBenchPredictionResponse(response_json=response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)
    
    
    def shutdown(self) -> None:
        """Represents the function to shutdown the predictor
           cleanup the endpooint/container/other resources
        """
        script = f"""#!/bin/sh

        {STOP_AND_RM_CONTAINER}
        """
        tmpdir = tempfile.gettempdir()
        script_file_path = os.path.join(tmpdir, "shutdown_container.sh")
        Path(script_file_path).write_text(script)
        st = os.stat(script_file_path)
        os.chmod(script_file_path, st.st_mode | stat.S_IEXEC)

        logger.info(f"going to run script {script_file_path}")
        subprocess.run(["bash", script_file_path], check=True)
        logger.info(f"done running bash script")
        return None
    
    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    # The rest ep is deployed on an instance that incurs hourly cost hence, the calculcate cost function
    # computes the cost of the experiment on an hourly basis. If your instance has a different pricing structure
    # modify this function.
    def calculate_cost(self,
                    instance_type: str,
                    instance_count: int,
                    pricing: Dict,
                    duration: float,
                    prompt_tokens: int,
                    completion_tokens: int) -> float:
        """Calculate the cost of each experiment run."""
        experiment_cost: Optional[float] = None
        try:
            instance_based_pricing = pricing['pricing']['instance_based']
            hourly_rate = instance_based_pricing.get(instance_type, None)
            logger.info(f"the hourly rate for running on {instance_type} is {hourly_rate}, instance_count={instance_count}")
            # calculating the experiment cost for instance based pricing
            instance_count = instance_count if instance_count else 1
            experiment_cost = (hourly_rate / 3600) * duration * instance_count
        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost
    
    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        # not implemented
        return None

    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return self._inference_spec.get("parameters")
    
    @property
    def platform_type(self) -> Dict:
        """The inference parameters property."""
        return constants.PLATFORM_EC2


def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    return ollamaPredictor(endpoint_name, inference_spec, metadata)
